# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import hashlib
import inspect
import os
from os.path import abspath, dirname, join
import re
import sys
import threading
import time

from .logger import logger
from .utils.disk import path_decode
from .utils.module import import_module
from .utils.parse import parse_list
from .utils.singleton import Singleton


class PluginManager(metaclass=Singleton):
    """Designed to be a singleton object to maintain all initialized loading of
    modules in memory."""

    # Description (used for logging)
    name = "Singleton Plugin"

    # Memory Space
    _id = "undefined"

    # Our Module Python path name
    module_name_prefix = f"apprise.{_id}"

    # The module path to scan
    module_path = join(abspath(dirname(__file__)), _id)

    # For filtering our result when scanning a module
    module_filter_re = re.compile(r"^(?P<name>((?!_)[A-Za-z0-9]+))$")

    # thread safe loading
    _lock = threading.Lock()

    def __init__(self, *args, **kwargs):
        """Over-ride our class instantiation to provide a singleton."""

        self._module_map = None
        self._schema_map = None

        # This contains a mapping of all plugins dynamicaly loaded at runtime
        # from external modules such as the @notify decorator
        #
        # The elements here will be additionally added to the _schema_map if
        # there is no conflict otherwise.
        # The structure looks like the following:
        # Module path, e.g. /usr/share/apprise/plugins/my_notify_hook.py
        # {
        #   'path': path,
        #
        #   'notify': {
        #     'schema': {
        #       'name': 'Custom schema name',
        #       'fn_name': 'name_of_function_decorator_was_found_on',
        #       'url': 'schema://any/additional/info/found/on/url'
        #       'plugin': <CustomNotifyWrapperPlugin>
        #    },
        #     'schema2': {
        #       'name': 'Custom schema name',
        #       'fn_name': 'name_of_function_decorator_was_found_on',
        #       'url': 'schema://any/additional/info/found/on/url'
        #       'plugin': <CustomNotifyWrapperPlugin>
        #    }
        #  }
        # Note: that the <CustomNotifyWrapperPlugin> inherits from
        #       NotifyBase
        self._custom_module_map = {}

        # Track manually disabled modules (by their schema)
        self._disabled = set()

        # Hash of all paths previously scanned so we don't waste
        # effort/overhead doing it again
        self._paths_previously_scanned = set()

        # Track loaded module paths to prevent from loading them again
        self._loaded = set()

    def unload_modules(self, disable_native=False):
        """Reset our object and unload all modules."""

        with self._lock:
            if self._custom_module_map:
                # Handle Custom Module Assignments
                for meta in self._custom_module_map.values():
                    if meta["name"] not in self._module_map:
                        # Nothing to remove
                        continue

                    # For the purpose of tidying up un-used modules in memory
                    loaded = [
                        m
                        for m in sys.modules
                        if m.startswith(self._module_map[meta["name"]]["path"])
                    ]

                    for module_path in loaded:
                        del sys.modules[module_path]

            # Reset disabled plugins (if any)
            for schema in self._disabled:
                self._schema_map[schema].enabled = True
            self._disabled.clear()

            # Reset our variables
            self._schema_map = {}
            self._custom_module_map = {}
            if disable_native:
                self._module_map = {}

            else:
                self._module_map = None
                self._loaded = set()

            # Reset our path cache
            self._paths_previously_scanned = set()

    def load_modules(self, path=None, name=None, force=False):
        """Load our modules into memory."""

        # Default value
        module_name_prefix = self.module_name_prefix if name is None else name
        module_path = self.module_path if path is None else path

        with self._lock:
            if not force and module_path in self._loaded:
                # We're done
                return

            # Our base reference
            module_count = len(self._module_map) if self._module_map else 0
            schema_count = len(self._schema_map) if self._schema_map else 0

            if not self:
                # Initialize our maps
                self._module_map = {}
                self._schema_map = {}
                self._custom_module_map = {}

            # Used for the detection of additional Notify Services objects
            # The .py extension is optional as we support loading directories
            # too
            module_re = re.compile(
                r"^(?P<name>(?!base|_)[a-z0-9_]+)(\.py)?$", re.I
            )

            t_start = time.time()
            for f in os.listdir(module_path):
                tl_start = time.time()
                match = module_re.match(f)
                if not match:
                    # keep going
                    continue

                # Store our notification/plugin name:
                module_name = match.group("name")
                module_pyname = f"{module_name_prefix}.{module_name}"

                if module_name in self._module_map:
                    logger.warning(
                        "%s(s) (%s) already loaded; ignoring %s",
                        self.name,
                        module_name,
                        os.path.join(module_path, f),
                    )
                    continue

                try:
                    module = __import__(
                        module_pyname,
                        globals(),
                        locals(),
                        fromlist=[module_name],
                    )

                except ImportError:
                    # No problem, we can try again another way...
                    module = import_module(
                        os.path.join(module_path, f), module_pyname
                    )
                    if not module:
                        # logging found in import_module and not needed here
                        continue

                module_class = None
                for m_class in [
                    obj
                    for obj in dir(module)
                    if self.module_filter_re.match(obj)
                ]:
                    # Get our plugin
                    plugin = getattr(module, m_class)
                    if not hasattr(plugin, "app_id"):
                        # Filter out non-notification modules
                        logger.trace(
                            "(%s.%s) import failed; no app_id defined in %s",
                            self.name,
                            m_class,
                            os.path.join(module_path, f),
                        )
                        continue

                    # Add our plugin name to our module map
                    self._module_map[module_name] = {
                        "plugin": {plugin},
                        "module": module,
                        "path": f"{module_name_prefix}.{module_name}",
                        "native": True,
                    }

                    fn = getattr(plugin, "schemas", None)
                    schemas = set() if not callable(fn) else fn(plugin)

                    # map our schema to our plugin
                    for schema in schemas:
                        if schema in self._schema_map:
                            logger.error(
                                f"{self.name} schema ({schema}) mismatch"
                                " detected -"
                                f" {self._schema_map[schema]} already maps to"
                                f" {plugin}"
                            )
                            continue

                        # Assign plugin
                        self._schema_map[schema] = plugin

                    # Store our class
                    module_class = m_class
                    break

                if not module_class:
                    # Not a library we can load as it doesn't follow the simple
                    # rule that the class must bear the same name as the
                    # notification file itself.
                    logger.trace(
                        "%s (%s) import failed; no filename/Class "
                        "match found in %s",
                        self.name,
                        module_name,
                        os.path.join(module_path, f),
                    )
                    continue

                logger.trace(
                    f"{self.name} {module_name} loaded in"
                    f" {time.time() - tl_start:.6f}s"
                )

            # Track the directory loaded so we never load it again
            self._loaded.add(module_path)

            logger.debug(
                f"{self.name} {len(self._module_map) - module_count}(s) and"
                f" {len(self._schema_map) - schema_count} Schema(s) loaded in"
                f" {time.time() - t_start:.4f}s"
            )

    def module_detection(self, paths, cache=True):
        """Leverage the @notify decorator and load all objects found matching
        this."""
        # A simple restriction that we don't allow periods in the filename at
        # all so it can't be hidden (Linux OS's) and it won't conflict with
        # Python path naming.  This also prevents us from loading any python
        # file that starts with an underscore or dash
        # We allow for __init__.py as well
        module_re = re.compile(
            r"^(?P<name>[_a-z0-9][a-z0-9._-]+)?(\.py)?$", re.I
        )

        # Validate if we're a loadable Python file or not
        valid_python_file_re = re.compile(r".+\.py(o|c)?$", re.IGNORECASE)

        if isinstance(paths, str):
            paths = [
                paths,
            ]

        if not paths or not isinstance(paths, (tuple, list)):
            # We're done
            return

        def _import_module(path):
            # Since our plugin name can conflict (as a module) with another
            # we want to generate random strings to avoid steping on
            # another's namespace
            if not (path and valid_python_file_re.match(path)):
                # Ignore file/module type
                logger.trace("Plugin Scan: Skipping %s", path)
                return

            t_start = time.time()
            module_name = hashlib.sha1(path.encode("utf-8")).hexdigest()
            module_pyname = "{prefix}.{name}".format(
                prefix="apprise.custom.module", name=module_name
            )

            if module_pyname in self._custom_module_map:
                # First clear out existing entries
                for schema in self._custom_module_map[module_pyname]["notify"]:

                    # Remove any mapped modules to this file
                    del self._schema_map[schema]

                # Reset
                del self._custom_module_map[module_pyname]

            # Load our module
            module = import_module(path, module_pyname)
            if not module:
                # No problem, we can't use this object
                logger.warning("Failed to load custom module: %s", _path)
                return

            # Print our loaded modules if any
            if module_pyname in self._custom_module_map:
                logger.debug(
                    "Custom module %s - %d schema(s) (name=%s) "
                    "loaded in %.6fs",
                    _path,
                    len(self._custom_module_map[module_pyname]["notify"]),
                    module_name,
                    (time.time() - t_start),
                )

                # Add our plugin name to our module map
                self._module_map[module_name] = {
                    "plugin": set(),
                    "module": module,
                    "path": module_pyname,
                    "native": False,
                }

                for schema, _meta in self._custom_module_map[module_pyname][
                    "notify"
                ].items():

                    # For mapping purposes; map our element in our main list
                    self._module_map[module_name]["plugin"].add(
                        self._schema_map[schema]
                    )

                    # Log our success
                    logger.info("Loaded custom notification: %s://", schema)
            else:
                # The code reaches here if we successfully loaded the Python
                # module but no hooks/triggers were found. So we can safely
                # just remove/ignore this entry
                del sys.modules[module_pyname]
                return

            # end of _import_module()
            return

        for _path in paths:
            path = path_decode(_path)
            if (
                cache and path in self._paths_previously_scanned
            ) or not os.path.exists(path):
                # We're done as we've already scanned this
                continue

            # Store our path as a way of hashing it has been handled
            self._paths_previously_scanned.add(path)

            if os.path.isdir(path) and not os.path.isfile(
                os.path.join(path, "__init__.py")
            ):

                logger.debug("Scanning for custom plugins in: %s", path)
                for entry in os.listdir(path):
                    re_match = module_re.match(entry)
                    if not re_match:
                        # keep going
                        logger.trace("Plugin Scan: Ignoring %s", entry)
                        continue

                    new_path = os.path.join(path, entry)
                    if os.path.isdir(new_path):
                        # Update our path
                        new_path = os.path.join(path, entry, "__init__.py")
                        if not os.path.isfile(new_path):
                            logger.trace(
                                "Plugin Scan: Ignoring %s",
                                os.path.join(path, entry),
                            )
                            continue

                    if not cache or (
                        cache
                        and new_path not in self._paths_previously_scanned
                    ):
                        # Load our module
                        _import_module(new_path)

                        # Add our subdir path
                        self._paths_previously_scanned.add(new_path)
            else:
                if os.path.isdir(path):
                    # This logic is safe to apply because we already
                    # validated the directories state above; update our
                    # path
                    path = os.path.join(path, "__init__.py")
                    if cache and path in self._paths_previously_scanned:
                        continue

                    self._paths_previously_scanned.add(path)

                # directly load as is
                re_match = module_re.match(os.path.basename(path))
                # must be a match and must have a .py extension
                if not re_match or not re_match.group(1):
                    # keep going
                    logger.trace("Plugin Scan: Ignoring %s", path)
                    continue

                # Load our module
                _import_module(path)

        return None

    def add(self, plugin, schemas=None, url=None, send_func=None):
        """Ability to manually add Notification services to our stack."""

        if not self:
            # Lazy load
            self.load_modules()

        # Acquire a list of schemas
        p_schemas = parse_list(plugin.secure_protocol, plugin.protocol)
        if isinstance(schemas, str):
            schemas = [
                schemas,
            ]

        elif schemas is None:
            # Default
            schemas = p_schemas

        if not schemas or not isinstance(schemas, (set, tuple, list)):
            # We're done
            logger.error(
                "The schemas provided (type %s) is unsupported; "
                "loaded from %s.",
                type(schemas),
                send_func.__name__ if send_func else plugin.__class__.__name__,
            )
            return False

        # Convert our schemas into a set
        schemas = {s.lower() for s in schemas} | set(p_schemas)

        # Valdation
        conflict = [s for s in schemas if s in self]
        if conflict:
            # we're already handling this schema
            logger.warning(
                "The schema(s) (%s) are already defined and could not be "
                "loaded from %s%s.",
                ", ".join(conflict),
                "custom notify function " if send_func else "",
                send_func.__name__ if send_func else plugin.__class__.__name__,
            )
            return False

        if send_func:
            # Acquire the function name
            fn_name = send_func.__name__

            # Acquire the python filename path
            path = inspect.getfile(send_func)

            # Acquire our path to our module
            module_name = str(send_func.__module__)

            if module_name not in self._custom_module_map:
                # Support non-dynamic includes as well...
                self._custom_module_map[module_name] = {
                    # Name can be useful for indexing back into the
                    # _module_map object; this is the key to do it with:
                    "name": module_name.split(".")[-1],
                    # The path to the module loaded
                    "path": path,
                    # Initialize our template
                    "notify": {},
                }

            for schema in schemas:
                self._custom_module_map[module_name]["notify"][schema] = {
                    # The name of the send function the @notify decorator
                    # wrapped
                    "fn_name": fn_name,
                    # The URL that was provided in the @notify decorator call
                    # associated with the 'on='
                    "url": url,
                }

        else:
            module_name = hashlib.sha1(
                "".join(schemas).encode("utf-8")
            ).hexdigest()
            module_pyname = "{prefix}.{name}".format(
                prefix="apprise.adhoc.module", name=module_name
            )

            # Add our plugin name to our module map
            self._module_map[module_name] = {
                "plugin": {plugin},
                "module": None,
                "path": module_pyname,
                "native": False,
            }

        for schema in schemas:
            # Assign our mapping
            self._schema_map[schema] = plugin

        return True

    def remove(self, *schemas):
        """Removes a loaded element (if defined)"""
        if not self:
            # Lazy load
            self.load_modules()

        for schema in schemas:
            with contextlib.suppress(KeyError):
                del self[schema]

    def plugins(self, include_disabled=True):
        """Return all of our loaded plugins."""
        if not self:
            # Lazy load
            self.load_modules()

        for module in self._module_map.values():
            for plugin in module["plugin"]:
                if not include_disabled and not plugin.enabled:
                    continue
                yield plugin

    def schemas(self, include_disabled=True):
        """Return all of our loaded schemas.

        if include_disabled == True, then even disabled notifications are
        returned
        """
        if not self:
            # Lazy load
            self.load_modules()

        # Return our list
        return (
            list(self._schema_map.keys())
            if include_disabled
            else [s for s in self._schema_map if self._schema_map[s].enabled]
        )

    def disable(self, *schemas):
        """Disables the modules associated with the specified schemas."""
        if not self:
            # Lazy load
            self.load_modules()

        for schema in schemas:
            if schema not in self._schema_map:
                continue

            if not self._schema_map[schema].enabled:
                continue

            # Disable
            self._schema_map[schema].enabled = False
            self._disabled.add(schema)

    def enable_only(self, *schemas):
        """Disables the modules associated with the specified schemas."""
        if not self:
            # Lazy load
            self.load_modules()

        # convert to set for faster indexing
        schemas = set(schemas)

        for plugin in self.plugins():
            # Get our plugin's schema list
            p_schemas = set(
                parse_list(plugin.secure_protocol, plugin.protocol)
            )

            if not schemas & p_schemas:
                if plugin.enabled:
                    # Disable it (only if previously enabled); this prevents us
                    # from adjusting schemas that were disabled due to missing
                    # libraries or other environment reasons
                    plugin.enabled = False
                    self._disabled |= p_schemas
                continue

            # If we reach here, our schema was flagged to be enabled
            if p_schemas & self._disabled:
                # Previously disabled; no worries, let's clear this up
                self._disabled -= p_schemas
                plugin.enabled = True

    def __contains__(self, schema):
        """Checks if a schema exists."""
        if not self:
            # Lazy load
            self.load_modules()

        return schema in self._schema_map

    def __delitem__(self, schema):
        if not self:
            # Lazy load
            self.load_modules()

        # Get our plugin (otherwise we throw a KeyError) which is
        # intended on del action that doesn't align
        plugin = self._schema_map[schema]

        # Our list of all schema entries
        p_schemas = {schema}

        for key in list(self._module_map.keys()):
            if plugin in self._module_map[key]["plugin"]:
                # Remove our plugin
                self._module_map[key]["plugin"].remove(plugin)

                # Custom Plugin Entry; Clean up cross reference
                module_pyname = self._module_map[key]["path"]
                if (
                    not self._module_map[key]["native"]
                    and module_pyname in self._custom_module_map
                ):

                    del self._custom_module_map[module_pyname]["notify"][
                        schema
                    ]

                    if not self._custom_module_map[module_pyname]["notify"]:
                        #
                        # Last custom loaded element
                        #

                        # Free up custom object entry
                        del self._custom_module_map[module_pyname]

                if not self._module_map[key]["plugin"]:
                    #
                    # Last element
                    #
                    if self._module_map[key]["native"]:
                        # Get our plugin's schema list
                        p_schemas = {
                            s
                            for s in parse_list(
                                plugin.secure_protocol, plugin.protocol
                            )
                            if s in self._schema_map
                        }

                    # free system memory
                    if self._module_map[key]["module"]:
                        del sys.modules[self._module_map[key]["path"]]

                    # free last remaining pointer in module map
                    del self._module_map[key]

        for schema in p_schemas:
            # Final Tidy
            del self._schema_map[schema]

    def __setitem__(self, schema, plugin):
        """Support fast assigning of Plugin/Notification Objects."""
        if not self:
            # Lazy load
            self.load_modules()

        # Set default values if not otherwise set
        if not plugin.service_name:
            # Assign service name if one doesn't exist
            plugin.service_name = f"{schema}://"

        p_schemas = set(parse_list(plugin.secure_protocol, plugin.protocol))
        if not p_schemas:
            # Assign our protocol
            plugin.secure_protocol = schema
            p_schemas.add(schema)

        elif schema not in p_schemas:
            # Add our others (if defined)
            plugin.secure_protocol = {
                schema,
                *parse_list(plugin.secure_protocol),
            }
            p_schemas.add(schema)

        if not self.add(plugin, schemas=p_schemas):
            raise KeyError("Conflicting Assignment")

    def __getitem__(self, schema):
        """Returns the indexed plugin identified by the schema specified."""
        if not self:
            # Lazy load
            self.load_modules()

        return self._schema_map[schema]

    def __iter__(self):
        """Returns an iterator so we can iterate over our loaded modules."""
        if not self:
            # Lazy load
            self.load_modules()

        return iter(self._module_map.values())

    def __len__(self):
        """Returns the number of modules/plugins loaded."""
        if not self:
            # Lazy load
            self.load_modules()

        return len(self._module_map)

    def __bool__(self):
        """Determines if object has loaded or not."""
        return bool(self._loaded and self._module_map is not None)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from os.path import abspath, dirname, join
import re

from .manager import PluginManager


class AttachmentManager(PluginManager):
    """Designed to be a singleton object to maintain all initialized attachment
    plugins/modules in memory."""

    # Description (used for logging)
    name = "Attachment Plugin"

    # Filename Prefix to filter on
    fname_prefix = "Attach"

    # Memory Space
    _id = "attachment"

    # Our Module Python path name
    module_name_prefix = f"apprise.{_id}"

    # The module path to scan
    module_path = join(abspath(dirname(__file__)), _id)

    # For filtering our result set
    module_filter_re = re.compile(
        r"^(?P<name>" + fname_prefix + r"(?!Base)[A-Za-z0-9]+)$"
    )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Information:
# - https://docs.aws.amazon.com/ses/latest/APIReference/API_SendRawEmail.html
#
# AWS Credentials (access_key and secret_access_key)
# - https://docs.aws.amazon.com/sdk-for-java/v1/developer-guide/\
#       setup-credentials.html
# - https://docs.aws.amazon.com/toolkit-for-eclipse/v1/user-guide/\
#       setup-credentials.html
#
#      Other systems write these credentials to:
#        -  ~/.aws/credentials on Linux, macOS, or Unix
#        -  C:\Users\USERNAME\.aws\credentials on Windows
#
#
#      To get A users access key ID and secret access key
#
#        1. Open the IAM console: https://console.aws.amazon.com/iam/home
#        2. On the navigation menu, choose Users.
#        3. Choose your IAM user name (not the check box).
#        4. Open the Security credentials tab, and then choose:
#             Create Access key - Programmatic access
#        5. To see the new access key, choose Show. Your credentials resemble
#           the following:
#               Access key ID: AKIAIOSFODNN7EXAMPLE
#               Secret access key: wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY
#
#      To download the key pair, choose Download .csv file. Store the keys
#      The account requries this permssion to 'SES v2 : SendEmail' in order to
#      work
#
#      To get the root users account (if you're logged in as that) you can
#      visit: https://console.aws.amazon.com/iam/home#/\
#                 security_credentials$access_key
#
#    This information is vital to work with SES


# To use/test the service, i logged into the portal via:
#       - https://portal.aws.amazon.com
#
# Go to the dashboard of the Amazon SES (Simple Email Service)
#  1. You must have a verified identity; click on that option and create one
#     if you don't already have one. Until it's verified, you won't be able to
#     do the next step.
#  2. From here you'll be able to retrieve your ARN associated with your
#     identity you want Apprise to send emails on behalf. It might look
#     something like:
#          arn:aws:ses:us-east-2:133216123003:identity/user@example.com
#
#  This is your ARN (Amazon Record Name)
#
#

import base64
from collections import OrderedDict
from datetime import datetime, timezone
from email.header import Header
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from hashlib import sha256
import hmac
import re
from urllib.parse import quote
from xml.etree import ElementTree

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_email, parse_emails, validate_regex
from .base import NotifyBase

# Our Regin Identifier
# support us-gov-west-1 syntax as well
IS_REGION = re.compile(
    r"^\s*(?P<country>[a-z]{2})-(?P<area>[a-z-]+?)-(?P<no>[0-9]+)\s*$", re.I
)

# Extend HTTP Error Messages
AWS_HTTP_ERROR_MAP = {
    403: "Unauthorized - Invalid Access/Secret Key Combination.",
}


class NotifySES(NotifyBase):
    """A wrapper for AWS SES (Amazon Simple Email Service)"""

    # The default descriptive name associated with the Notification
    service_name = "AWS Simple Email Service (SES)"

    # The services URL
    service_url = "https://aws.amazon.com/ses/"

    # The default secure protocol
    secure_protocol = "ses"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_ses"

    # Support attachments
    attachment_support = True

    # AWS is pretty good for handling data load so request limits
    # can occur in much shorter bursts
    request_rate_per_sec = 2.5

    # Default Notify Format
    notify_format = NotifyFormat.HTML

    # Define object templates
    templates = (
        (
            "{schema}://{from_email}/{access_key_id}/{secret_access_key}/"
            "{region}/{targets}"
        ),
        "{schema}://{from_email}/{access_key_id}/{secret_access_key}/{region}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "from_email": {
                "name": _("From Email"),
                "type": "string",
                "map_to": "from_addr",
                "required": True,
            },
            "access_key_id": {
                "name": _("Access Key ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "secret_access_key": {
                "name": _("Secret Access Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "region": {
                "name": _("Region"),
                "type": "string",
                "regex": (r"^[a-z]{2}-[a-z-]+?-[0-9]+$", "i"),
                "required": True,
                "map_to": "region_name",
            },
            "targets": {
                "name": _("Target Emails"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_email",
            },
            "reply": {
                "name": _("Reply To Email"),
                "type": "string",
                "map_to": "reply_to",
            },
            "name": {
                "name": _("From Name"),
                "type": "string",
                "map_to": "from_name",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
            "access": {
                "alias_of": "access_key_id",
            },
            "secret": {
                "alias_of": "secret_access_key",
            },
            "region": {
                "alias_of": "region",
            },
        },
    )

    def __init__(
        self,
        access_key_id,
        secret_access_key,
        region_name,
        reply_to=None,
        from_addr=None,
        from_name=None,
        targets=None,
        cc=None,
        bcc=None,
        **kwargs,
    ):
        """Initialize Notify AWS SES Object."""
        super().__init__(**kwargs)

        # Store our AWS API Access Key
        self.aws_access_key_id = validate_regex(access_key_id)
        if not self.aws_access_key_id:
            msg = "An invalid AWS Access Key ID was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our AWS API Secret Access key
        self.aws_secret_access_key = validate_regex(secret_access_key)
        if not self.aws_secret_access_key:
            msg = (
                "An invalid AWS Secret Access Key "
                f"({secret_access_key}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire our AWS Region Name:
        # eg. us-east-1, cn-north-1, us-west-2, ...
        self.aws_region_name = validate_regex(
            region_name, *self.template_tokens["region"]["regex"]
        )
        if not self.aws_region_name:
            msg = f"An invalid AWS Region ({region_name}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire Email 'To'
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # For tracking our email -> name lookups
        self.names = {}

        # Set our notify_url based on our region
        self.notify_url = f"https://email.{self.aws_region_name}.amazonaws.com"

        # AWS Service Details
        self.aws_service_name = "ses"
        self.aws_canonical_uri = "/"

        # AWS Authentication Details
        self.aws_auth_version = "AWS4"
        self.aws_auth_algorithm = "AWS4-HMAC-SHA256"
        self.aws_auth_request = "aws4_request"

        # Get our From username (if specified)
        self.from_name = from_name

        if from_addr:
            self.from_addr = from_addr

        else:
            # Get our from email address
            self.from_addr = f"{self.user}@{self.host}" if self.user else None

        if not (self.from_addr and is_email(self.from_addr)):
            msg = "An invalid AWS From ({}) was specified.".format(
                f"{self.user}@{self.host}"
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.reply_to = None
        if reply_to:
            result = is_email(reply_to)
            if not result:
                msg = "An invalid AWS Reply To ({}) was specified.".format(
                    f"{reply_to}"
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            self.reply_to = (
                result["name"] if result["name"] else False,
                result["full_email"],
            )

        if targets:
            # Validate recipients (to:) and drop bad ones:
            for recipient in parse_emails(targets):
                result = is_email(recipient)
                if result:
                    self.targets.append((
                        result["name"] if result["name"] else False,
                        result["full_email"],
                    ))
                    continue

                self.logger.warning(
                    f"Dropped invalid To email ({recipient}) specified.",
                )

        else:
            # If our target email list is empty we want to add ourselves to it
            self.targets.append(
                (self.from_name if self.from_name else False, self.from_addr)
            )

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            email = is_email(recipient)
            if email:
                self.cc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                f"Dropped invalid Carbon Copy email ({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):
            email = is_email(recipient)
            if email:
                self.bcc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                f"({recipient}) specified.",
            )

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Wrapper to send_notification since we can alert more then one
        channel."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning("There are no SES email recipients to notify")
            return False

        # error tracking (used for function return)
        has_error = False

        # Initialize our default from name
        from_name = (
            self.from_name
            if self.from_name
            else (
                self.reply_to[0]
                if self.reply_to and self.reply_to[0]
                else self.app_desc
            )
        )

        reply_to = (
            from_name,
            self.from_addr if not self.reply_to else self.reply_to[1],
        )

        # Create a copy of the targets list
        emails = list(self.targets)
        while len(emails):
            # Get our email to notify
            to_name, to_addr = emails.pop(0)

            # Strip target out of cc list if in To or Bcc
            cc = self.cc - self.bcc - {to_addr}

            # Strip target out of bcc list if in To
            bcc = self.bcc - {to_addr}

            # Format our cc addresses to support the Name field
            cc = [
                formataddr(
                    (self.names.get(addr, False), addr), charset="utf-8"
                )
                for addr in cc
            ]

            # Format our bcc addresses to support the Name field
            bcc = [
                formataddr(
                    (self.names.get(addr, False), addr), charset="utf-8"
                )
                for addr in bcc
            ]

            self.logger.debug(
                "Email From: {} <{}>".format(
                    quote(reply_to[0], " "), quote(reply_to[1], "@ ")
                )
            )

            self.logger.debug(f"Email To: {to_addr}")
            if cc:
                self.logger.debug("Email Cc: {}".format(", ".join(cc)))
            if bcc:
                self.logger.debug("Email Bcc: {}".format(", ".join(bcc)))

            # Prepare Email Message
            if self.notify_format == NotifyFormat.HTML:
                content = MIMEText(body, "html", "utf-8")

            else:
                content = MIMEText(body, "plain", "utf-8")

            # Create a Multipart container if there is an attachment
            base = (
                MIMEMultipart()
                if attach and self.attachment_support
                else content
            )

            # TODO: Deduplicate with `NotifyEmail`?
            base["Subject"] = Header(title, "utf-8")
            base["From"] = formataddr(
                (from_name if from_name else False, self.from_addr),
                charset="utf-8",
            )
            base["To"] = formataddr((to_name, to_addr), charset="utf-8")
            if reply_to[1] != self.from_addr:
                base["Reply-To"] = formataddr(reply_to, charset="utf-8")
            base["Cc"] = ",".join(cc)
            base["Date"] = datetime.now(timezone.utc).strftime(
                "%a, %d %b %Y %H:%M:%S +0000"
            )
            base["X-Application"] = self.app_id

            if attach and self.attachment_support:
                # First attach our body to our content as the first element
                base.attach(content)

                # Now store our attachments
                for no, attachment in enumerate(attach, start=1):
                    if not attachment:
                        # We could not load the attachment; take an early
                        # exit since this isn't what the end user wanted

                        # We could not access the attachment
                        self.logger.error(
                            "Could not access attachment"
                            f" {attachment.url(privacy=True)}."
                        )

                        return False

                    self.logger.debug(
                        "Preparing Email attachment"
                        f" {attachment.url(privacy=True)}"
                    )

                    with open(attachment.path, "rb") as abody:
                        app = MIMEApplication(abody.read())
                        app.set_type(attachment.mimetype)

                        filename = (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        )

                        app.add_header(
                            "Content-Disposition",
                            'attachment; filename="{}"'.format(
                                Header(filename, "utf-8")
                            ),
                        )

                        base.attach(app)

            # Prepare our payload object
            payload = {
                "Action": "SendRawEmail",
                "Version": "2010-12-01",
                "RawMessage.Data": (
                    base64.b64encode(base.as_string().encode("utf-8")).decode(
                        "utf-8"
                    )
                ),
            }

            for no, email in enumerate(([to_addr, *bcc, *cc]), start=1):
                payload[f"Destinations.member.{no}"] = email

            # Specify from address
            payload["Source"] = "{} <{}>".format(
                quote(from_name, " "), quote(self.from_addr, "@ ")
            )

            (result, _response) = self._post(payload=payload, to=to_addr)
            if not result:
                # Mark our failure
                has_error = True
                continue

        return not has_error

    def _post(self, payload, to):
        """Wrapper to request.post() to manage it's response better and make
        the send() function cleaner and easier to maintain.

        This function returns True if the _post was successful and False if it
        wasn't.
        """

        # Always call throttle before any remote server i/o is made; for AWS
        # time plays a huge factor in the headers being sent with the payload.
        # So for AWS (SES) requests we must throttle before they're generated
        # and not directly before the i/o call like other notification
        # services do.
        self.throttle()

        # Convert our payload from a dict() into a urlencoded string
        payload = NotifySES.urlencode(payload)

        # Prepare our Notification URL
        # Prepare our AWS Headers based on our payload
        headers = self.aws_prepare_request(payload)

        self.logger.debug(
            "AWS SES POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug("AWS SES Payload (%d bytes)", len(payload))

        try:
            r = requests.post(
                self.notify_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifySES.http_response_code_lookup(
                    r.status_code, AWS_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send AWS SES notification to {}: "
                    "{}{}error={}.".format(
                        to,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return (False, NotifySES.aws_response_to_dict(r.text))

            else:
                self.logger.info(f'Sent AWS SES notification to "{to}".')

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending AWS SES "
                f'notification to "{to}".',
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return (False, NotifySES.aws_response_to_dict(None))

        return (True, NotifySES.aws_response_to_dict(r.text))

    def aws_prepare_request(self, payload, reference=None):
        """Takes the intended payload and returns the headers for it.

        The payload is presumed to have been already urlencoded()
        """

        # Define our AWS SES header
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
            # Populated below
            "Content-Length": 0,
            "Authorization": None,
            "X-Amz-Date": None,
        }

        # Get a reference time (used for header construction)
        reference = datetime.now(timezone.utc)

        # Provide Content-Length
        headers["Content-Length"] = str(len(payload))

        # Amazon Date Format
        amzdate = reference.strftime("%Y%m%dT%H%M%SZ")
        headers["X-Amz-Date"] = amzdate

        # Credential Scope
        scope = "{date}/{region}/{service}/{request}".format(
            date=reference.strftime("%Y%m%d"),
            region=self.aws_region_name,
            service=self.aws_service_name,
            request=self.aws_auth_request,
        )

        # Similar to headers; but a subset.  keys must be lowercase
        signed_headers = OrderedDict([
            ("content-type", headers["Content-Type"]),
            ("host", f"email.{self.aws_region_name}.amazonaws.com"),
            ("x-amz-date", headers["X-Amz-Date"]),
        ])

        #
        # Build Canonical Request Object
        #
        canonical_request = "\n".join([
            # Method
            "POST",
            # URL
            self.aws_canonical_uri,
            # Query String (none set for POST)
            "",
            # Header Content (must include \n at end!)
            # All entries except characters in amazon date must be
            # lowercase
            "\n".join([f"{k}:{v}" for k, v in signed_headers.items()]) + "\n",
            # Header Entries (in same order identified above)
            ";".join(signed_headers.keys()),
            # Payload
            sha256(payload.encode("utf-8")).hexdigest(),
        ])

        # Prepare Unsigned Signature
        to_sign = "\n".join([
            self.aws_auth_algorithm,
            amzdate,
            scope,
            sha256(canonical_request.encode("utf-8")).hexdigest(),
        ])

        # Our Authorization header
        headers["Authorization"] = ", ".join([
            (
                f"{self.aws_auth_algorithm} "
                f"Credential={self.aws_access_key_id}/{scope}"
            ),
            "SignedHeaders={signed_headers}".format(
                signed_headers=";".join(signed_headers.keys()),
            ),
            f"Signature={self.aws_auth_signature(to_sign, reference)}",
        ])

        return headers

    def aws_auth_signature(self, to_sign, reference):
        """Generates a AWS v4 signature based on provided payload which should
        be in the form of a string."""

        def _sign(key, msg, to_hex=False):
            """Perform AWS Signing."""
            if to_hex:
                return hmac.new(key, msg.encode("utf-8"), sha256).hexdigest()
            return hmac.new(key, msg.encode("utf-8"), sha256).digest()

        _date = _sign(
            (self.aws_auth_version + self.aws_secret_access_key).encode(
                "utf-8"
            ),
            reference.strftime("%Y%m%d"),
        )

        _region = _sign(_date, self.aws_region_name)
        _service = _sign(_region, self.aws_service_name)
        _signed = _sign(_service, self.aws_auth_request)
        return _sign(_signed, to_sign, to_hex=True)

    @staticmethod
    def aws_response_to_dict(aws_response):
        """Takes an AWS Response object as input and returns it as a dictionary
        but not befor extracting out what is useful to us first.

        eg:
          IN:

            <SendRawEmailResponse
                 xmlns="http://ses.amazonaws.com/doc/2010-12-01/">
              <SendRawEmailResult>
                <MessageId>
                   010f017d87656ee2-a2ea291f-79ea-
                   44f3-9d25-00d041de3007-000000</MessageId>
              </SendRawEmailResult>
              <ResponseMetadata>
                <RequestId>7abb454e-904b-4e46-a23c-2f4d2fc127a6</RequestId>
              </ResponseMetadata>
            </SendRawEmailResponse>

          OUT:
           {
             'type': 'SendRawEmailResponse',
              'message_id': '010f017d87656ee2-a2ea291f-79ea-
                             44f3-9d25-00d041de3007-000000',
              'request_id': '7abb454e-904b-4e46-a23c-2f4d2fc127a6',
           }
        """

        # Define ourselves a set of directives we want to keep if found and
        # then identify the value we want to map them to in our response
        # object
        aws_keep_map = {
            "RequestId": "request_id",
            "MessageId": "message_id",
            # Error Message Handling
            "Type": "error_type",
            "Code": "error_code",
            "Message": "error_message",
        }

        # A default response object that we'll manipulate as we pull more data
        # from our AWS Response object
        response = {
            "type": None,
            "request_id": None,
            "message_id": None,
        }

        try:
            # we build our tree, but not before first eliminating any
            # reference to namespacing (if present) as it makes parsing
            # the tree so much easier.
            root = ElementTree.fromstring(
                re.sub(' xmlns="[^"]+"', "", aws_response, count=1)
            )

            # Store our response tag object name
            response["type"] = str(root.tag)

            def _xml_iter(root, response):
                if len(root) > 0:
                    for child in root:
                        # use recursion to parse everything
                        _xml_iter(child, response)

                elif root.tag in aws_keep_map:
                    response[aws_keep_map[root.tag]] = (root.text).strip()

            # Recursivly iterate over our AWS Response to extract the
            # fields we're interested in in efforts to populate our response
            # object.
            _xml_iter(root, response)

        except (ElementTree.ParseError, TypeError):
            # bad data just causes us to generate a bad response
            pass

        return response

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.from_addr,
            self.aws_access_key_id,
            self.aws_secret_access_key,
            self.aws_region_name,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Acquire any global URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.from_name is not None:
            # from_name specified; pass it back on the url
            params["name"] = self.from_name

        if self.cc:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                "{}{}".format(
                    "" if not e not in self.names else f"{self.names[e]}:",
                    e,
                )
                for e in self.cc
            ])

        if self.bcc:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(self.bcc)

        if self.reply_to:
            # Handle our reply to address
            params["reply"] = (
                "{} <{}>".format(*self.reply_to)
                if self.reply_to[0]
                else self.reply_to[1]
            )

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = not (
            len(self.targets) == 1 and self.targets[0][1] == self.from_addr
        )

        return (
            "{schema}://{from_addr}/{key_id}/{key_secret}/{region}/"
            "{targets}/?{params}".format(
                schema=self.secure_protocol,
                from_addr=NotifySES.quote(self.from_addr, safe="@"),
                key_id=self.pprint(self.aws_access_key_id, privacy, safe=""),
                key_secret=self.pprint(
                    self.aws_secret_access_key,
                    privacy,
                    mode=PrivacyMode.Secret,
                    safe="",
                ),
                region=NotifySES.quote(self.aws_region_name, safe=""),
                targets=(
                    ""
                    if not has_targets
                    else "/".join([
                        NotifySES.quote(
                            "{}{}".format(
                                "" if not e[0] else f"{e[0]}:", e[1]
                            ),
                            safe="",
                        )
                        for e in self.targets
                    ])
                ),
                params=NotifySES.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        entries = NotifySES.split_path(results["fullpath"])

        # The AWS Access Key ID is stored in the first entry
        access_key_id = entries.pop(0) if entries else None

        # Our AWS Access Key Secret contains slashes in it which unfortunately
        # means it is of variable length after the hostname.  Since we require
        # that the user provides the region code, we intentionally use this
        # as our delimiter to detect where our Secret is.
        secret_access_key = None
        region_name = None

        # We need to iterate over each entry in the fullpath and find our
        # region. Once we get there we stop and build our secret from our
        # accumulated data.
        secret_access_key_parts = []

        # Section 1: Get Region and Access Secret
        index = 0
        for index, entry in enumerate(entries, start=1):

            # Are we at the region yet?
            result = IS_REGION.match(entry)
            if result:
                # Ensure region is nicely formatted
                region_name = "{country}-{area}-{no}".format(
                    country=result.group("country").lower(),
                    area=result.group("area").lower(),
                    no=result.group("no"),
                )

                # We're done with Section 1 of our url (the credentials)
                break

            elif is_email(entry):
                # We're done with Section 1 of our url (the credentials)
                index -= 1
                break

            # Store our secret parts
            secret_access_key_parts.append(entry)

        # Prepare our Secret Access Key
        secret_access_key = (
            "/".join(secret_access_key_parts)
            if secret_access_key_parts
            else None
        )

        # Section 2: Get our Recipients (basically all remaining entries)
        results["targets"] = entries[index:]

        if "name" in results["qsd"] and len(results["qsd"]["name"]):
            # Extract from name to associate with from address
            results["from_name"] = NotifySES.unquote(results["qsd"]["name"])

        # Handle 'to' email address
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(results["qsd"]["to"])

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = NotifySES.parse_list(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = NotifySES.parse_list(results["qsd"]["bcc"])

        # Handle From Address handling
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["from_addr"] = NotifySES.unquote(results["qsd"]["from"])

        # Handle Reply To Address
        if "reply" in results["qsd"] and len(results["qsd"]["reply"]):
            results["reply_to"] = NotifySES.unquote(results["qsd"]["reply"])

        # Handle secret_access_key over-ride
        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            results["secret_access_key"] = NotifySES.unquote(
                results["qsd"]["secret"]
            )
        else:
            results["secret_access_key"] = secret_access_key

        # Handle access key id over-ride
        if "access" in results["qsd"] and len(results["qsd"]["access"]):
            results["access_key_id"] = NotifySES.unquote(
                results["qsd"]["access"]
            )
        else:
            results["access_key_id"] = access_key_id

        # Handle region name id over-ride
        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            results["region_name"] = NotifySES.unquote(
                results["qsd"]["region"]
            )
        else:
            results["region_name"] = region_name

        # Return our result set
        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# All of the documentation needed to work with the Twist API can be found
# here: https://developer.twist.com/v3/

from itertools import chain
from json import loads
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_email, parse_list
from .base import NotifyBase

# A workspace can also be interpreted as a team name too!
IS_CHANNEL = re.compile(
    r"^#?(?P<name>((?P<workspace>[A-Za-z0-9_-]+):)?"
    r"(?P<channel>[^\s]{1,64}))$"
)

IS_CHANNEL_ID = re.compile(
    r"^(?P<name>((?P<workspace>[0-9]+):)?(?P<channel>[0-9]+))$"
)

# Used to break apart list of potential tags by their delimiter
# into a usable list.
LIST_DELIM = re.compile(r"[ \t\r\n,\\/]+")


class NotifyTwist(NotifyBase):
    """A wrapper for Notify Twist Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Twist"

    # The services URL
    service_url = "https://twist.com"

    # The default secure protocol
    secure_protocol = "twist"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_twist"

    # The maximum size of the message
    body_maxlen = 1000

    # Default to markdown
    notify_format = NotifyFormat.MARKDOWN

    # The default Notification URL to use
    api_url = "https://api.twist.com/api/v3/"

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # The default channel to notify if no targets are specified
    default_notification_channel = "general"

    # Define object templates
    templates = (
        "{schema}://{password}:{email}",
        "{schema}://{password}:{email}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "email": {
                "name": _("Email"),
                "type": "string",
                "required": True,
            },
            "target_channel": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "target_channel_id": {
                "name": _("Target Channel ID"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(self, email=None, targets=None, **kwargs):
        """Initialize Notify Twist Object."""
        super().__init__(**kwargs)

        # Initialize channels list
        self.channels = set()

        # Initialize Channel ID which are stored as:
        #   <workspace_id>:<channel_id>
        self.channel_ids = set()

        # The token is None if we're not logged in and False if we
        # failed to log in.  Otherwise it is set to the actual token
        self.token = None

        # Our default workspace (associated with our token)
        self.default_workspace = None

        # A set of all of the available workspaces
        self._cached_workspaces = set()

        # A mapping of channel names, the layout is as follows:
        #  {
        #     <workspace_id>: {
        #          <channel_name>: <channel_id>,
        #          <channel_name>: <channel_id>,
        #          ...
        #     },
        #     <workspace2_id>: {
        #          <channel_name>: <channel_id>,
        #          <channel_name>: <channel_id>,
        #          ...
        #     },
        #  }
        self._cached_channels = {}

        # Initialize our Email Object
        self.email = email if email else f"{self.user}@{self.host}"

        # Check if it is valid
        result = is_email(self.email)
        if not result:
            # let outer exception handle this
            msg = f"The Twist Auth email specified ({self.email}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Re-assign email based on what was parsed
        self.email = result["full_email"]
        if email:
            # Force user/host to be that of the defined email for
            # consistency. This is very important for those initializing
            # this object with the the email object would could potentially
            # cause inconsistency to contents in the NotifyBase() object
            self.user = result["user"]
            self.host = result["domain"]

        if not self.password:
            msg = f"No Twist password was specified with account: {self.email}"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate recipients and drop bad ones:
        for recipient in parse_list(targets):
            result = IS_CHANNEL_ID.match(recipient)
            if result:
                # store valid channel id
                self.channel_ids.add(result.group("name"))
                continue

            result = IS_CHANNEL.match(recipient)
            if result:
                # store valid device
                self.channels.add(result.group("name").lower())
                continue

            self.logger.warning(
                f"Dropped invalid channel/id ({recipient}) specified.",
            )

        if len(self.channels) + len(self.channel_ids) == 0:
            # Notify our default channel
            self.channels.add(self.default_notification_channel)
            self.logger.warning(
                "Added default notification channel "
                f"{self.default_notification_channel}"
            )
        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return (
            "{schema}://{password}:{user}@{host}/{targets}/?{params}".format(
                schema=self.secure_protocol,
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                user=self.quote(self.user, safe=""),
                host=self.host,
                targets="/".join([
                    NotifyTwist.quote(x, safe="")
                    for x in chain(
                        # Channels are prefixed with a pound/hashtag symbol
                        [f"#{x}" for x in self.channels],
                        # Channel IDs
                        self.channel_ids,
                    )
                ]),
                params=NotifyTwist.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.channels) + len(self.channel_ids)

    def login(self):
        """A simple wrapper to authenticate with the Twist Server."""

        # Prepare our payload
        payload = {
            "email": self.email,
            "password": self.password,
        }

        # Reset our default workspace
        self.default_workspace = None

        # Reset our cached objects
        self._cached_workspaces = set()
        self._cached_channels = {}

        # Send Login Information
        postokay, response = self._fetch(
            "users/login",
            payload=payload,
            # We set this boolean so internal recursion doesn't take place.
            login=True,
        )

        if not postokay or not response:
            # Setting this variable to False as a way of letting us know
            # we failed to authenticate on our last attempt
            self.token = False
            return False

        # Our response object looks like this (content has been altered for
        # presentation purposes):
        # {
        #     "contact_info": null,
        #     "profession": null,
        #     "timezone": "UTC",
        #     "avatar_id": null,
        #     "id": 123456,
        #     "first_name": "Jordan",
        #     "comet_channel":
        #         "124371-34be423219130343030d4ec0a3dabbbbbe565eee",
        #     "restricted": false,
        #     "default_workspace": 92020,
        #     "snooze_dnd_end": null,
        #     "email": "user@example.com",
        #     "comet_server": "https://comet.twist.com",
        #     "snooze_until": null,
        #     "lang": "en",
        #     "feature_flags": [],
        #     "short_name": "Jordan P.",
        #     "away_mode": null,
        #     "time_format": "12",
        #     "client_id": "cb01f37e-a5b2-13e9-ba2a-023a33d10dc0",
        #     "removed": false,
        #     "emails": [
        #         {
        #             "connected": [],
        #             "email": "user@example.com",
        #             "primary": true
        #         }
        #     ],
        #     "scheduled_banners": [
        #         "threads_3",
        #         "threads_1",
        #         "notification_permissions",
        #         "search_1",
        #         "messages_1",
        #         "team_1",
        #         "inbox_2",
        #         "inbox_1"
        #     ],
        #     "snooze_dnd_start": null,
        #     "name": "Jordan Peterson",
        #     "off_days": [],
        #     "bot": false,
        #     "token": "2e82c1e4e8b0091fdaa34ff3972351821406f796",
        #     "snoozed": false,
        #     "setup_pending": false,
        #     "date_format": "MM/DD/YYYY"
        # }

        # Store our default workspace
        self.default_workspace = response.get("default_workspace")

        # Acquire our token
        self.token = response.get("token")

        self.logger.info(f"Authenticated to Twist as {self.email}")
        return True

    def logout(self):
        """A simple wrapper to log out of the server."""

        if not self.token:
            # Nothing more to do
            return True

        # Send Logout Message
        _postokay, _response = self._fetch("users/logout")

        # reset our token
        self.token = None

        # There is no need to handling failed log out attempts at this time
        return True

    def get_workspaces(self):
        """Returns all workspaces associated with this user account as a set.

        This returned object is either an empty dictionary or one that
        looks like this:
           {
             'workspace': <workspace_id>,
             'workspace': <workspace_id>,
             'workspace': <workspace_id>,
           }

        All workspaces are made lowercase for comparison purposes
        """
        if not self.token and not self.login():
            # Nothing more to do
            return {}

        postokay, response = self._fetch("workspaces/get")
        if not postokay or not response:
            # We failed to retrieve
            return {}

        # The response object looks like so:
        #   [
        #     {
        #       "created_ts": 1563044447,
        #       "name": "apprise",
        #       "creator": 123571,
        #       "color": 1,
        #       "default_channel": 13245,
        #       "plan": "free",
        #       "default_conversation": 63022,
        #       "id": 12345
        #     }
        #   ]

        # Knowing our response, we can iterate over each object and cache our
        # object
        result = {}
        for entry in response:
            result[entry.get("name", "").lower()] = entry.get("id", "")

        return result

    def get_channels(self, wid):
        """Simply returns the channel objects associated with the specified
        workspace id.

        This returned object is either an empty dictionary or one that
        looks like this:
           {
             'channel1': <channel_id>,
             'channel2': <channel_id>,
             'channel3': <channel_id>,
           }

        All channels are made lowercase for comparison purposes
        """
        if not self.token and not self.login():
            # Nothing more to do
            return {}

        payload = {"workspace_id": wid}
        postokay, response = self._fetch("channels/get", payload=payload)

        if not postokay or not isinstance(response, list):
            # We failed to retrieve
            return {}

        # Response looks like this:
        #  [
        #    {
        #      "id": 123,
        #      "name": "General"
        #      "workspace_id": 12345,
        #      "color": 1,
        #      "description": "",
        #      "archived": false,
        #      "public": true,
        #      "user_ids": [
        #        8754
        #      ],
        #      "created_ts": 1563044447,
        #      "creator": 123571,
        #    }
        #  ]
        #
        # Knowing our response, we can iterate over each object and cache our
        # object
        result = {}
        for entry in response:
            result[entry.get("name", "").lower()] = entry.get("id", "")

        return result

    def _channel_migration(self):
        """A simple wrapper to get all of the current workspaces including the
        default one.  This plays a role in what channel(s) get notified and
        where.

        A cache lookup has overhead, and is only required to be preformed if
        the user specified channels by their string value
        """

        if not self.token and not self.login():
            # Nothing more to do
            return False

        if not len(self.channels):
            # Nothing to do; take an early exit
            return True

        if (
            self.default_workspace
            and self.default_workspace not in self._cached_channels
        ):
            # Get our default workspace entries
            self._cached_channels[self.default_workspace] = self.get_channels(
                self.default_workspace
            )

        # initialize our error tracking
        has_error = False

        while len(self.channels):
            # Pop our channel off of the stack
            result = IS_CHANNEL.match(self.channels.pop())

            # Populate our key variables
            workspace = result.group("workspace")
            channel = result.group("channel").lower()

            # Acquire our workspace_id if we can
            if workspace:
                # We always work with the workspace in it's lowercase form
                workspace = workspace.lower()

                # A workspace was defined
                if not len(self._cached_workspaces):
                    # cache our workspaces; this only needs to be done once
                    self._cached_workspaces = self.get_workspaces()

                if workspace not in self._cached_workspaces:
                    # not found
                    self.logger.warning(
                        f"The Twist User {self.email} is not associated "
                        f"with the Team {workspace}"
                    )

                    # Toggle our return flag
                    has_error = True
                    continue

                # Store the workspace id
                workspace_id = self._cached_workspaces[workspace]

            else:
                # use default workspace
                workspace_id = self.default_workspace

            # Check to see if our channel exists in our default workspace
            if (
                workspace_id in self._cached_channels
                and channel in self._cached_channels[workspace_id]
            ):
                # Store our channel ID
                self.channel_ids.add(
                    f"{workspace_id}"
                    f":{self._cached_channels[workspace_id][channel]}"
                )
                continue

            # if we reach here, we failed to add our channel
            self.logger.warning(
                "The Channel #{} was not found{}.".format(
                    channel,
                    "" if not workspace else f" with Team {workspace}",
                )
            )

            # Toggle our return flag
            has_error = True
            continue

        # There is no need to handling failed log out attempts at this time
        return not has_error

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Twist Notification."""

        # error tracking (used for function return)
        has_error = False

        if not self.token and not self.login():
            # We failed to authenticate - we're done
            return False

        if len(self.channels) > 0:
            # Converts channels to their maped IDs if found; this is the only
            # way to send notifications to Twist
            self._channel_migration()

        if not len(self.channel_ids):
            # We have nothing to notify
            self.logger.warning("There are no Twist targets to notify")
            return False

        # Notify all of our identified channels
        ids = list(self.channel_ids)
        while len(ids) > 0:
            # Retrieve our Channel Object
            result = IS_CHANNEL_ID.match(ids.pop())

            # We need both the workspace/team id and channel id
            channel_id = int(result.group("channel"))

            # Prepare our payload
            payload = {
                "channel_id": channel_id,
                "title": title,
                "content": body,
            }

            postokay, _response = self._fetch(
                "threads/add",
                payload=payload,
            )

            # only toggle has_error flag if we had an error
            if not postokay:
                # Mark our failure
                has_error = True
                continue

            # If we reach here, we were successful
            self.logger.info(
                "Sent Twist notification to {}.".format(result.group("name"))
            )

        return not has_error

    def _fetch(self, url, payload=None, method="POST", login=False):
        """Wrapper to Twist API requests object."""

        # use what was specified, otherwise build headers dynamically
        headers = {
            "User-Agent": self.app_id,
        }

        headers["Content-Type"] = (
            "application/x-www-form-urlencoded; charset=utf-8"
        )

        if self.token:
            # Set our token
            headers["Authorization"] = f"Bearer {self.token}"

        # Prepare our api url
        api_url = f"{self.api_url}{url}"

        # Some Debug Logging
        self.logger.debug(
            f"Twist {method} URL: {api_url} "
            f"(cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"Twist Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made;
        self.throttle()

        # Initialize a default value for our content value
        content = {}

        # acquire our request mode
        fn = requests.post if method == "POST" else requests.get
        try:
            r = fn(
                api_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # Get our JSON content if it's possible
            try:
                content = loads(r.content)

            except (TypeError, ValueError, AttributeError):
                # TypeError = r.content is not a String
                # ValueError = r.content is Unparsable
                # AttributeError = r.content is None
                content = {}

            # handle authentication errors where our token has just simply
            # expired. The error response content looks like this:
            #  {
            #     "error_code": 200,
            #     "error_uuid": "af80bd0715434231a649f2258d7fb946",
            #     "error_extra": {},
            #     "error_string": "Invalid token"
            #  }
            #
            #  Authentication related codes:
            #    120 = You are not logged in
            #    200 = Invalid Token
            #
            #  Source: https://developer.twist.com/v3/#errors
            #
            #  We attempt to login again and retry the original request
            #  if we aren't in the process of handling a login already
            if (
                r.status_code != requests.codes.ok
                and login is False
                and isinstance(content, dict)
                and content.get("error_code") in (120, 200)
                and self.login()
            ):

                r = fn(
                    api_url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # Get our JSON content if it's possible
                try:
                    content = loads(r.content)

                except (TypeError, ValueError, AttributeError):
                    # TypeError = r.content is not a String
                    # ValueError = r.content is Unparsable
                    # AttributeError = r.content is None
                    content = {}

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyTwist.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Twist {} to {}: {}error={}.".format(
                        method,
                        api_url,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

        except requests.RequestException as e:
            self.logger.warning(
                f"Exception received when sending Twist {method} to {api_url}"
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return (False, content)

        return (True, content)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if not results.get("user"):
            # A username is required
            return None

        # Acquire our targets
        results["targets"] = NotifyTwist.split_path(results["fullpath"])

        if not results.get("password"):
            # Password is required; we will accept the very first entry on the
            # path as a password instead
            if len(results["targets"]) == 0:
                # No targets to get our password from
                return None

            # We need to requote contents since this variable will get
            # unquoted later on in the process.  This step appears a bit
            # hacky, but it allows us to support the password in this location
            #   - twist://user@example.com/password
            results["password"] = NotifyTwist.quote(
                results["targets"].pop(0), safe=""
            )

        else:
            # Now we handle our format:
            #    twist://password:email
            #
            # since URL logic expects
            #    schema://user:password@host
            #
            # you can see how this breaks. The colon at the front delmits
            #  passwords and you can see the twist:// url inverts what we
            #  expect:
            #    twist://password:user@example.com
            #
            # twist://abc123:bob@example.com using normal conventions would
            # have interpreted 'bob' as the password and 'abc123' as the user.
            # For the purpose of apprise simplifying this for us, we need to
            # swap these arguments when we prepare the email.

            _password = results["user"]
            results["user"] = results["password"]
            results["password"] = _password

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyTwist.parse_list(results["qsd"]["to"])

        return results

    def __del__(self):
        """Destructor."""
        self.logout()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Youll need your own Revolt Bot and a Channel Id for the notifications to
# be sent in since Revolt does not support webhooks yet.
#
#  This plugin will simply work using the url of:
#     revolt://BOT_TOKEN/CHANNEL_ID
#
# API Documentation:
#    - https://api.revolt.chat/swagger/index.html
#

from datetime import datetime, timedelta, timezone
from json import dumps, loads

import requests

from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase


class NotifyRevolt(NotifyBase):
    """A wrapper for Revolt Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Revolt"

    # The services URL
    service_url = "https://revolt.chat/"

    # The default secure protocol
    secure_protocol = "revolt"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_revolt"

    # Revolt Channel Message
    notify_url = "https://api.revolt.chat/"

    # Revolt supports attachments but doesn't support it here (for now)
    attachment_support = False

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_256

    # Revolt is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # X-RateLimit-Reset: The epoc time (in seconds) we can expect our
    #                    rate-limit to be reset.
    # X-RateLimit-Remaining: an integer identifying how many requests we're
    #                        still allow to make.
    request_rate_per_sec = 3

    # Safety net
    clock_skew = timedelta(seconds=2)

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 2000

    # Title Maximum Length
    title_maxlen = 100

    # Define object templates
    templates = ("{schema}://{bot_token}/{targets}",)

    # Defile out template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "bot_token": {
                "name": _("Bot Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_channel": {
                "name": _("Channel ID"),
                "type": "string",
                "map_to": "targets",
                "regex": (r"^[a-z0-9_-]+$", "i"),
                "private": True,
                "required": True,
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "channel": {
                "alias_of": "targets",
            },
            "bot_token": {
                "alias_of": "bot_token",
            },
            "icon_url": {"name": _("Icon URL"), "type": "string"},
            "url": {
                "name": _("Embed URL"),
                "type": "string",
                "map_to": "link",
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(self, bot_token, targets, icon_url=None, link=None, **kwargs):
        super().__init__(**kwargs)

        # Bot Token
        self.bot_token = validate_regex(bot_token)
        if not self.bot_token:
            msg = f"An invalid Revolt Bot Token ({bot_token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Parse our Channel IDs
        self.targets = []
        for target in parse_list(targets):
            results = validate_regex(
                target, *self.template_tokens["target_channel"]["regex"]
            )

            if not results:
                self.logger.warning(
                    f"Dropped invalid Revolt channel ({target}) specified.",
                )
                continue

            # Add our target
            self.targets.append(target)

        # Image for Embed
        self.icon_url = icon_url

        # Url for embed title
        self.link = link

        # For Tracking Purposes
        self.ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

        # Default to 1.0
        self.ratelimit_remaining = 1.0

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Revolt Notification."""

        if len(self.targets) == 0:
            self.logger.warning("There were not Revolt channels to notify.")
            return False

        payload = {}

        # Acquire image_url
        image_url = (
            self.icon_url if self.icon_url else self.image_url(notify_type)
        )

        if self.notify_format == NotifyFormat.MARKDOWN:
            payload["embeds"] = [{
                "title": None if not title else title[0 : self.title_maxlen],
                "description": body,
                # Our color associated with our notification
                "colour": self.color(notify_type),
                "replies": None,
            }]

            if image_url:
                payload["embeds"][0]["icon_url"] = image_url

            if self.link:
                payload["embeds"][0]["url"] = self.link

        else:
            payload["content"] = body if not title else f"{title}\n{body}"

        has_error = False
        channel_ids = list(self.targets)
        for channel_id in channel_ids:
            postokay, _response = self._send(payload, channel_id)
            if not postokay:
                # Failed to send message
                has_error = True

        return not has_error

    def _send(self, payload, channel_id, retries=1, **kwargs):
        """Wrapper to the requests (post) object."""

        headers = {
            "User-Agent": self.app_id,
            "X-Bot-Token": self.bot_token,
            "Content-Type": "application/json; charset=utf-8",
            "Accept": "application/json; charset=utf-8",
        }

        notify_url = f"{self.notify_url}channels/{channel_id}/messages"

        self.logger.debug(
            "Revolt POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Revolt Payload: {payload!s}")

        # By default set wait to None
        wait = None

        now = datetime.now(timezone.utc).replace(tzinfo=None)
        if self.ratelimit_remaining <= 0.0 and now < self.ratelimit_reset:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Discord server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly and set our throttle
            # accordingly
            wait = abs(
                (
                    self.ratelimit_reset - now + self.clock_skew
                ).total_seconds()
            )

        # Default content response object
        content = {}

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                content = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            # Handle rate limiting (if specified)
            try:
                # Store our rate limiting (if provided)
                self.ratelimit_remaining = int(
                    r.headers.get("X-RateLimit-Remaining")
                )
                self.ratelimit_reset = now + timedelta(
                    seconds=(
                        int(r.headers.get("X-RateLimit-Reset-After")) / 1000
                    )
                )

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this
                # information gracefully accept this state and move on
                pass

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):

                # Some details to debug by
                self.logger.debug(
                    f"Response Details:\r\n{content if content else r.content}"
                )

                # We had a problem
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Revolt request limit reached; "
                    "instructed to throttle for %.3fs",
                    abs(
                        (
                            self.ratelimit_reset - now + self.clock_skew
                        ).total_seconds()
                    ),
                )

                if (
                    r.status_code == requests.codes.too_many_requests
                    and retries > 0
                ):

                    # Try again
                    return self._send(
                        payload=payload,
                        channel_id=channel_id,
                        retries=retries - 1,
                        **kwargs,
                    )

                self.logger.warning(
                    "Failed to send to Revolt notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                # Return; we're done
                return (False, content)

            else:
                self.logger.info("Sent Revolt notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred posting to Revolt."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return (False, content)

        return (True, content)

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.bot_token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {}

        if self.icon_url:
            params["icon_url"] = self.icon_url

        if self.link:
            params["url"] = self.link

        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{bot_token}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            bot_token=self.pprint(self.bot_token, privacy, safe=""),
            targets="/".join(
                [self.pprint(x, privacy, safe="") for x in self.targets]
            ),
            params=NotifyRevolt.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return 1 if not self.targets else len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our bot token
        bot_token = NotifyRevolt.unquote(results["host"])

        # Now fetch the Channel IDs
        targets = NotifyRevolt.split_path(results["fullpath"])

        results["bot_token"] = bot_token
        results["targets"] = targets

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyRevolt.parse_list(results["qsd"]["to"])

        # Support channel id on the URL string (if specified)
        if "channel" in results["qsd"]:
            results["targets"] += NotifyRevolt.parse_list(
                results["qsd"]["channel"]
            )

        # Support bot token on the URL string (if specified)
        if "bot_token" in results["qsd"]:
            results["bot_token"] = NotifyRevolt.unquote(
                results["qsd"]["bot_token"]
            )

        if "icon_url" in results["qsd"]:
            results["icon_url"] = NotifyRevolt.unquote(
                results["qsd"]["icon_url"]
            )

        if "url" in results["qsd"]:
            results["link"] = NotifyRevolt.unquote(results["qsd"]["url"])

        if "format" not in results["qsd"] and (
            "url" in results or "icon_url" in results
        ):
            # Markdown is implied
            results["format"] = NotifyFormat.MARKDOWN

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# 1. Visit https://www.reddit.com/prefs/apps and scroll to the bottom
# 2. Click on the button that reads 'are you a developer? create an app...'
# 3. Set the mode to `script`,
# 4. Provide a `name`, `description`, `redirect uri` and save it.
# 5. Once the bot is saved, you'll be given a ID (next to the the bot name)
#    and a Secret.

# The App ID will look something like this: YWARPXajkk645m
# The App Secret will look something like this: YZGKc5YNjq3BsC-bf7oBKalBMeb1xA
# The App will also have a location where you can identify the users
# who have access (identified as Developers) to the app itself. You will
# additionally need these credentials authenticate with.

# With this information you'll be able to form the URL:
# reddit://{user}:{password}@{app_id}/{app_secret}

# All of the documentation needed to work with the Reddit API can be found
# here:
#   - https://www.reddit.com/dev/api/
#   - https://www.reddit.com/dev/api/#POST_api_submit
#   - https://github.com/reddit-archive/reddit/wiki/API
from datetime import datetime, timedelta, timezone
from json import loads

import requests

from .. import __title__, __version__
from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
REDDIT_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token",
}


class RedditMessageKind:
    """Define the kinds of messages supported."""

    # Attempt to auto-detect the type prior to passing along the message to
    # Reddit
    AUTO = "auto"

    # A common message
    SELF = "self"

    # A Hyperlink
    LINK = "link"


REDDIT_MESSAGE_KINDS = (
    RedditMessageKind.AUTO,
    RedditMessageKind.SELF,
    RedditMessageKind.LINK,
)


class NotifyReddit(NotifyBase):
    """A wrapper for Notify Reddit Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Reddit"

    # The services URL
    service_url = "https://reddit.com"

    # The default secure protocol
    secure_protocol = "reddit"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_reddit"

    # The maximum size of the message
    body_maxlen = 6000

    # Maximum title length as defined by the Reddit API
    title_maxlen = 300

    # Default to markdown
    notify_format = NotifyFormat.MARKDOWN

    # The default Notification URL to use
    auth_url = "https://www.reddit.com/api/v1/access_token"
    submit_url = "https://oauth.reddit.com/api/submit"

    # Reddit is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # X-RateLimit-Reset: The epoc time (in seconds) we can expect our
    #                    rate-limit to be reset.
    # X-RateLimit-Remaining: an integer identifying how many requests we're
    #                        still allow to make.
    request_rate_per_sec = 0

    # Taken right from google.auth.helpers:
    clock_skew = timedelta(seconds=10)

    # 1 hour in seconds (the lifetime of our token)
    access_token_lifetime_sec = timedelta(seconds=3600)

    # Define object templates
    templates = (
        "{schema}://{user}:{password}@{app_id}/{app_secret}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "app_id": {
                "name": _("Application ID"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
            "app_secret": {
                "name": _("Application Secret"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
            "target_subreddit": {
                "name": _("Target Subreddit"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "kind": {
                "name": _("Kind"),
                "type": "choice:string",
                "values": REDDIT_MESSAGE_KINDS,
                "default": RedditMessageKind.AUTO,
            },
            "flair_id": {
                "name": _("Flair ID"),
                "type": "string",
                "map_to": "flair_id",
            },
            "flair_text": {
                "name": _("Flair Text"),
                "type": "string",
                "map_to": "flair_text",
            },
            "nsfw": {
                "name": _("NSFW"),
                "type": "bool",
                "default": False,
                "map_to": "nsfw",
            },
            "ad": {
                "name": _("Is Ad?"),
                "type": "bool",
                "default": False,
                "map_to": "advertisement",
            },
            "replies": {
                "name": _("Send Replies"),
                "type": "bool",
                "default": True,
                "map_to": "sendreplies",
            },
            "spoiler": {
                "name": _("Is Spoiler"),
                "type": "bool",
                "default": False,
                "map_to": "spoiler",
            },
            "resubmit": {
                "name": _("Resubmit Flag"),
                "type": "bool",
                "default": False,
                "map_to": "resubmit",
            },
        },
    )

    def __init__(
        self,
        app_id=None,
        app_secret=None,
        targets=None,
        kind=None,
        nsfw=False,
        sendreplies=True,
        resubmit=False,
        spoiler=False,
        advertisement=False,
        flair_id=None,
        flair_text=None,
        **kwargs,
    ):
        """Initialize Notify Reddit Object."""
        super().__init__(**kwargs)

        # Initialize subreddit list
        self.subreddits = set()

        # Not Safe For Work Flag
        self.nsfw = nsfw

        # Send Replies Flag
        self.sendreplies = sendreplies

        # Is Spoiler Flag
        self.spoiler = spoiler

        # Resubmit Flag
        self.resubmit = resubmit

        # Is Ad?
        self.advertisement = advertisement

        # Flair details
        self.flair_id = flair_id
        self.flair_text = flair_text

        # Our keys we build using the provided content
        self.__refresh_token = None
        self.__access_token = None
        self.__access_token_expiry = datetime.now(timezone.utc)

        self.kind = (
            kind.strip().lower()
            if isinstance(kind, str)
            else self.template_args["kind"]["default"]
        )

        if self.kind not in REDDIT_MESSAGE_KINDS:
            msg = f"An invalid Reddit message kind ({kind}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        self.user = validate_regex(self.user)
        if not self.user:
            msg = f"An invalid Reddit User ID ({self.user}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        self.password = validate_regex(self.password)
        if not self.password:
            msg = f"An invalid Reddit Password ({self.password}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        self.client_id = validate_regex(
            app_id, *self.template_tokens["app_id"]["regex"]
        )
        if not self.client_id:
            msg = f"An invalid Reddit App ID ({app_id}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        self.client_secret = validate_regex(
            app_secret, *self.template_tokens["app_secret"]["regex"]
        )
        if not self.client_secret:
            msg = f"An invalid Reddit App Secret ({app_secret}) was specified"
            self.logger.warning(msg)
            raise TypeError(msg)

        # Build list of subreddits
        self.subreddits = [
            sr.lstrip("#") for sr in parse_list(targets) if sr.lstrip("#")
        ]

        if not self.subreddits:
            self.logger.warning("No subreddits were identified to be notified")

        # For Rate Limit Tracking Purposes
        self.ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

        # Default to 1.0
        self.ratelimit_remaining = 1.0

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.client_id,
            self.client_secret,
            self.user,
            self.password,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "kind": self.kind,
            "ad": "yes" if self.advertisement else "no",
            "nsfw": "yes" if self.nsfw else "no",
            "resubmit": "yes" if self.resubmit else "no",
            "replies": "yes" if self.sendreplies else "no",
            "spoiler": "yes" if self.spoiler else "no",
        }

        # Flair support
        if self.flair_id:
            params["flair_id"] = self.flair_id

        if self.flair_text:
            params["flair_text"] = self.flair_text

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{user}:{password}@{app_id}/{app_secret}"
            "/{targets}/?{params}".format(
                schema=self.secure_protocol,
                user=NotifyReddit.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                app_id=self.pprint(
                    self.client_id, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                app_secret=self.pprint(
                    self.client_secret,
                    privacy,
                    mode=PrivacyMode.Secret,
                    safe="",
                ),
                targets="/".join(
                    [NotifyReddit.quote(x, safe="") for x in self.subreddits]
                ),
                params=NotifyReddit.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.subreddits)

    def login(self):
        """A simple wrapper to authenticate with the Reddit Server."""

        # Prepare our payload
        payload = {
            "grant_type": "password",
            "username": self.user,
            "password": self.password,
        }

        # Enforce a False flag setting before calling _fetch()
        self.__access_token = False

        # Send Login Information
        postokay, response = self._fetch(
            self.auth_url,
            payload=payload,
        )

        if not postokay or not response:
            # Setting this variable to False as a way of letting us know
            # we failed to authenticate on our last attempt
            self.__access_token = False
            return False

        # Our response object looks like this (content has been altered for
        # presentation purposes):
        # {
        #     "access_token": Your access token,
        #     "token_type": "bearer",
        #     "expires_in": Unix Epoch Seconds,
        #     "scope": A scope string,
        #     "refresh_token": Your refresh token
        # }

        # Acquire our token
        self.__access_token = response.get("access_token")

        # Handle other optional arguments we can use
        if "expires_in" in response:
            delta = timedelta(seconds=int(response["expires_in"]))
            self.__access_token_expiry = (
                delta + datetime.now(timezone.utc) - self.clock_skew
            )
        else:
            self.__access_token_expiry = (
                self.access_token_lifetime_sec
                + datetime.now(timezone.utc)
                - self.clock_skew
            )

        # The Refresh Token
        self.__refresh_token = response.get(
            "refresh_token", self.__refresh_token
        )

        if self.__access_token:
            self.logger.info(f"Authenticated to Reddit as {self.user}")
            return True

        self.logger.warning(f"Failed to authenticate to Reddit as {self.user}")

        # Mark our failure
        return False

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Reddit Notification."""

        # error tracking (used for function return)
        has_error = False

        if not self.__access_token and not self.login():
            # We failed to authenticate - we're done
            return False

        if not len(self.subreddits):
            # We have nothing to notify; we're done
            self.logger.warning("There are no Reddit targets to notify")
            return False

        # Prepare our Message Type/Kind
        if self.kind == RedditMessageKind.AUTO:
            parsed = NotifyBase.parse_url(body)
            # Detect a link
            if (
                parsed
                and parsed.get("schema", "").startswith("http")
                and parsed.get("host")
            ):
                kind = RedditMessageKind.LINK

            else:
                kind = RedditMessageKind.SELF
        else:
            kind = self.kind

        # Create a copy of the subreddits list
        subreddits = list(self.subreddits)
        while len(subreddits) > 0:
            # Retrieve our subreddit
            subreddit = subreddits.pop()

            # Prepare our payload
            payload = {
                "ad": bool(self.advertisement),
                "api_type": "json",
                "extension": "json",
                "sr": subreddit,
                "title": title if title else self.app_desc,
                "kind": kind,
                "nsfw": bool(self.nsfw),
                "resubmit": bool(self.resubmit),
                "sendreplies": bool(self.sendreplies),
                "spoiler": bool(self.spoiler),
            }

            if self.flair_id:
                payload["flair_id"] = self.flair_id

            if self.flair_text:
                payload["flair_text"] = self.flair_text

            if kind == RedditMessageKind.LINK:
                payload.update({
                    "url": body,
                })
            else:
                payload.update({
                    "text": body,
                })

            postokay, _response = self._fetch(self.submit_url, payload=payload)
            # only toggle has_error flag if we had an error
            if not postokay:
                # Mark our failure
                has_error = True
                continue

            # If we reach here, we were successful
            self.logger.info(f"Sent Reddit notification to {subreddit}")

        return not has_error

    def _fetch(self, url, payload=None):
        """Wrapper to Reddit API requests object."""

        # use what was specified, otherwise build headers dynamically
        headers = {"User-Agent": f"{__title__} v{__version__}"}

        if self.__access_token:
            # Set our token
            headers["Authorization"] = f"Bearer {self.__access_token}"

        # Prepare our url
        url = self.submit_url if self.__access_token else self.auth_url

        # Some Debug Logging
        self.logger.debug(
            f"Reddit POST URL: {url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"Reddit Payload: {payload!s}")

        # By default set wait to None
        wait = None

        if self.ratelimit_remaining <= 0.0:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Reddit server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly:

            now = datetime.now(timezone.utc).replace(tzinfo=None)
            if now < self.ratelimit_reset:
                # We need to throttle for the difference in seconds
                wait = abs(
                    (
                        self.ratelimit_reset - now + self.clock_skew
                    ).total_seconds()
                )

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        # Initialize a default value for our content value
        content = {}

        # acquire our request mode
        try:
            r = requests.post(
                url,
                data=payload,
                auth=(
                    None
                    if self.__access_token
                    else (self.client_id, self.client_secret)
                ),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            #  We attempt to login again and retry the original request
            #  if we aren't in the process of handling a login already
            if (
                r.status_code != requests.codes.ok
                and self.__access_token
                and url != self.auth_url
            ):

                # We had a problem
                status_str = NotifyReddit.http_response_code_lookup(
                    r.status_code, REDDIT_HTTP_ERROR_MAP
                )

                self.logger.debug(
                    "Taking countermeasures after failed to send to Reddit "
                    "{}: {}error={}".format(
                        url, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # We failed to authenticate with our token; login one more
                # time and retry this original request
                if not self.login():
                    return (False, {})

                # Try again
                r = requests.post(
                    url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

            # Get our JSON content if it's possible
            try:
                content = loads(r.content)

            except (TypeError, ValueError, AttributeError):
                # TypeError = r.content is not a String
                # ValueError = r.content is Unparsable
                # AttributeError = r.content is None

                # We had a problem
                status_str = NotifyReddit.http_response_code_lookup(
                    r.status_code, REDDIT_HTTP_ERROR_MAP
                )

                # Reddit always returns a JSON response
                self.logger.warning(
                    "Failed to send to Reddit after countermeasures {}: "
                    "{}error={}".format(
                        url, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return (False, {})

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyReddit.http_response_code_lookup(
                    r.status_code, REDDIT_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send to Reddit {}: {}error={}".format(
                        url, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

            errors = (
                []
                if not content
                else content.get("json", {}).get("errors", [])
            )
            if errors:
                self.logger.warning(
                    f"Failed to send to Reddit {url}: {errors!s}"
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

            try:
                # Store our rate limiting (if provided)
                self.ratelimit_remaining = float(
                    r.headers.get("X-RateLimit-Remaining")
                )
                self.ratelimit_reset = datetime.fromtimestamp(
                    int(r.headers.get("X-RateLimit-Reset")), timezone.utc
                ).replace(tzinfo=None)

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this information
                # gracefully accept this state and move on
                pass

        except requests.RequestException as e:
            self.logger.warning(
                f"Exception received when sending Reddit to {url}"
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return (False, content)

        return (True, content)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Acquire our targets
        results["targets"] = NotifyReddit.split_path(results["fullpath"])

        # Kind override
        if "kind" in results["qsd"] and results["qsd"]["kind"]:
            results["kind"] = NotifyReddit.unquote(
                results["qsd"]["kind"].strip().lower()
            )
        else:
            results["kind"] = RedditMessageKind.AUTO

        # Is an Ad?
        results["ad"] = parse_bool(results["qsd"].get("ad", False))

        # Get Not Safe For Work (NSFW) Flag
        results["nsfw"] = parse_bool(results["qsd"].get("nsfw", False))

        # Send Replies Flag
        results["replies"] = parse_bool(results["qsd"].get("replies", True))

        # Resubmit Flag
        results["resubmit"] = parse_bool(results["qsd"].get("resubmit", False))

        # Is Spoiler Flag
        results["spoiler"] = parse_bool(results["qsd"].get("spoiler", False))

        if "flair_text" in results["qsd"]:
            results["flair_text"] = NotifyReddit.unquote(
                results["qsd"]["flair_text"]
            )

        if "flair_id" in results["qsd"]:
            results["flair_id"] = NotifyReddit.unquote(
                results["qsd"]["flair_id"]
            )

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyReddit.parse_list(results["qsd"]["to"])

        if "app_id" in results["qsd"]:
            results["app_id"] = NotifyReddit.unquote(results["qsd"]["app_id"])
        else:
            # The App/Bot ID is the hostname
            results["app_id"] = NotifyReddit.unquote(results["host"])

        if "app_secret" in results["qsd"]:
            results["app_secret"] = NotifyReddit.unquote(
                results["qsd"]["app_secret"]
            )
        else:
            # The first target identified is the App secret
            results["app_secret"] = (
                None if not results["targets"] else results["targets"].pop(0)
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you need to create a MS Teams Azure Webhook Workflow:
#  https://support.microsoft.com/en-us/office/browse-and-add-workflows-\
#       in-microsoft-teams-4998095c-8b72-4b0e-984c-f2ad39e6ba9a

# Your webhook will look somthing like this (legacy):
# https://prod-161.westeurope.logic.azure.com:443/\
#       workflows/643e69f83c8944438d68119179a10a64/triggers/manual/\
#       paths/invoke?api-version=2016-06-01&sp=%2Ftriggers%2Fmanual%2Frun&\
#       sv=1.0&sig=KODuebWbDGYFr0z0eu-6Rj8aUKz7108W3wrNJZxFE5A
#
# Or it may now look something like this:
# https://prod-161.westeurope.logic.azure.com:443/\
#       powerautomate/automations/direct/\
#       workflows/643e69f83c8944438d68119179a10a64/triggers/manual/\
#       paths/invoke?api-version=2022-03-01-preview&sp=%2Ftriggers%2Fmanual%2F\
#       run&sv=1.0&sig=KODuebWbDGYFr0z0eu-6Rj8aUKz7108W3wrNJZxFE5A
#
# Yes... The URL is that big... But it looks like this (greatly simplified):
# https://HOST:PORT/workflows/ABCD/triggers/manual/path/...sig=DEFG
#          ^    ^                ^                              ^
#          |    |                |                              |
#  These are important <---------^------------------------------^
#
#
# Apprise can support this webhook as is (directly passed into it)
# Alternatively it can be shortend to:

# These 3 tokens need to be placed in the URL after the Team
#   workflows://HOST:PORT/ABCD/DEFG/
#

import json
from json.decoder import JSONDecodeError
import re

import requests

from ..apprise_attachment import AppriseAttachment
from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, validate_regex
from ..utils.templates import TemplateType, apply_template
from .base import NotifyBase


class APIVersion:
    """
    Define API Versions
    """
    WORKFLOW = "2016-06-01"
    POWER_AUTOMATE = "2022-03-01-preview"


class NotifyWorkflows(NotifyBase):
    """A wrapper for Microsoft Workflows (MS Teams) Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Power Automate / Workflows (for MSTeams)"

    # The services URL
    service_url = (
        "https://www.microsoft.com/power-platform/products/power-automate"
    )

    # The default secure protocol
    secure_protocol = ("workflow", "workflows")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_workflows"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_32

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1000

    # Default Notification Format
    notify_format = NotifyFormat.MARKDOWN

    # There is no reason we should exceed 35KB when reading in a JSON file.
    # If it is more than this, then it is not accepted
    max_workflows_template_size = 35000

    # Adaptive Card Version
    adaptive_card_version = "1.4"

    # Define object templates
    templates = (
        "{schema}://{host}/{workflow}/{signature}",
        "{schema}://{host}:{port}/{workflow}/{signature}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            # workflow identifier
            "workflow": {
                "name": _("Workflow ID"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
            # Signature
            "signature": {
                "name": _("Signature"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "id": {
                "alias_of": "workflow",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "pa": {
                "name": _("Use Power Automate URL"),
                "type": "bool",
                "default": False,
                "map_to": "power_automate",
            },
            "powerautomate": {"alias_of": "pa"},
            "wrap": {
                "name": _("Wrap Text"),
                "type": "bool",
                "default": True,
            },
            "template": {
                "name": _("Template Path"),
                "type": "string",
                "private": True,
            },
            # Below variable shortforms are taken from the Workflows webhook
            # for consistency
            "sig": {
                "alias_of": "signature",
            },
            "ver": {
                "name": _("API Version"),
                "type": "string",
                "map_to": "version",
            },
            "api-version": {"alias_of": "ver"},
        },
    )

    # Define our token control
    template_kwargs = {
        "tokens": {
            "name": _("Template Tokens"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        workflow,
        signature,
        include_image=None,
        power_automate=None,
        version=None,
        template=None,
        tokens=None,
        wrap=None,
        **kwargs,
    ):
        """Initialize Microsoft Workflows Object."""
        super().__init__(**kwargs)

        self.workflow = validate_regex(
            workflow, *self.template_tokens["workflow"]["regex"]
        )
        if not self.workflow:
            msg = f"An invalid Workflows ID ({workflow}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.signature = validate_regex(
            signature, *self.template_tokens["signature"]["regex"]
        )
        if not self.signature:
            msg = f"An invalid Signature ({signature}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Place a thumbnail image inline with the message body
        self.include_image = bool(
            include_image
            if include_image is not None
            else self.template_args["image"]["default"]
        )

        # Power Automate status
        self.power_automate = bool(
            power_automate
            if power_automate is not None
            else self.template_args["pa"]["default"]
        )

        # Wrap Text
        self.wrap = bool(
            wrap if wrap is not None else self.template_args["wrap"]["default"]
        )

        # Our template object is just an AppriseAttachment object
        self.template = AppriseAttachment(asset=self.asset)
        if template:
            # Add our definition to our template
            self.template.add(template)
            # Enforce maximum file size
            self.template[0].max_file_size = self.max_workflows_template_size

        # Prepare Version
        # The default is taken from the template_args
        # - If using power_automate, the API version required is different.
        default_api_version = (
            APIVersion.POWER_AUTOMATE
            if self.power_automate else APIVersion.WORKFLOW)

        self.api_version = (
            version
            if version is not None
            else default_api_version
        )

        # Template functionality
        self.tokens = {}
        if isinstance(tokens, dict):
            self.tokens.update(tokens)

        elif tokens:
            msg = (
                "The specified Workflows Template Tokens "
                f"({tokens}) are not identified as a dictionary."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # else:  NoneType - this is okay
        return

    def gen_payload(
        self, body, title="", notify_type=NotifyType.INFO, **kwargs
    ):
        """This function generates our payload whether it be the generic one
        Apprise generates by default, or one provided by a specified external
        template."""

        # Acquire our to-be footer icon if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        body_content = []
        if image_url:
            body_content.append({
                "type": "Image",
                "url": image_url,
                "height": "32px",
                "altText": notify_type,
            })

        if title:
            body_content.append({
                "type": "TextBlock",
                "text": f"{title}",
                "style": "heading",
                "weight": "Bolder",
                "size": "Large",
                "id": "title",
            })

        body_content.append({
            "type": "TextBlock",
            "text": body,
            "style": "default",
            "wrap": self.wrap,
            "id": "body",
        })

        if not self.template:
            # By default we use a generic working payload if there was
            # no template specified
            schema = "http://adaptivecards.io/schemas/adaptive-card.json"
            payload = {
                "type": "message",
                "attachments": [{
                    "contentType": "application/vnd.microsoft.card.adaptive",
                    "contentUrl": None,
                    "content": {
                        "$schema": schema,
                        "type": "AdaptiveCard",
                        "version": self.adaptive_card_version,
                        "body": body_content,
                        # Additionally
                        "msteams": {"width": "full"},
                    },
                }],
            }

            return payload

        # If our code reaches here, then we generate ourselves the payload
        template = self.template[0]
        if not template:
            # We could not access the attachment
            self.logger.error(
                "Could not access Workflow template"
                f" {template.url(privacy=True)}."
            )
            return False

        # Take a copy of our token dictionary
        tokens = self.tokens.copy()

        # Apply some defaults template values
        tokens["app_body"] = body
        tokens["app_title"] = title
        tokens["app_type"] = notify_type.value
        tokens["app_id"] = self.app_id
        tokens["app_desc"] = self.app_desc
        tokens["app_color"] = self.color(notify_type)
        tokens["app_image_url"] = image_url
        tokens["app_url"] = self.app_url

        # Enforce Application mode
        tokens["app_mode"] = TemplateType.JSON

        try:
            with open(template.path) as fp:
                content = json.loads(apply_template(fp.read(), **tokens))

        except OSError:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} could not be"
                " read."
            )
            return None

        except JSONDecodeError as e:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} contains"
                " invalid JSON."
            )
            self.logger.debug(f"JSONDecodeError: {e}")
            return None

        return content

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Microsoft Teams Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        params = {
            "api-version": self.api_version,
            "sp": "/triggers/manual/run",
            "sv": "1.0",
            "sig": self.signature,
        }

        # The URL changes depending on whether we're using power automate or
        # not
        path = (
            "/powerautomate/automations/direct"
            if self.power_automate
            else ""
        )

        notify_url = (
            "https://{host}{port}{path}/workflows/{workflow}/"
            "triggers/manual/paths/invoke".format(
                host=self.host,
                port="" if not self.port else f":{self.port}",
                path=path,
                workflow=self.workflow,
            )
        )

        # Generate our payload if it's possible
        payload = self.gen_payload(
            body=body, title=title, notify_type=notify_type, **kwargs
        )
        if not payload:
            # No need to present a reason; that will come from the
            # gen_payload() function itself
            return False

        self.logger.debug(
            "Workflows POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Workflows Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                notify_url,
                params=params,
                data=json.dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.accepted,
            ):
                # We had a problem
                status_str = NotifyWorkflows.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Workflows notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # We failed
                return False

            else:
                self.logger.info("Sent Workflows notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Workflows notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # We failed
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol[0],
            self.host,
            self.port,
            self.workflow,
            self.signature,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
            "wrap": "yes" if self.wrap else "no",
            "pa": "yes" if self.power_automate else "no",
        }

        if self.template:
            params["template"] = NotifyWorkflows.quote(
                self.template[0].url(), safe=""
            )

        # Store our version if it differs from default
        if (self.api_version != APIVersion.WORKFLOW
            and not self.power_automate) or (
                self.api_version != APIVersion.POWER_AUTOMATE
                and self.power_automate):
            # But only do so if we're not using power automate with the
            # default version for that.
            params["ver"] = self.api_version

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))
        # Store any template entries if specified
        params.update({f":{k}": v for k, v in self.tokens.items()})

        return (
            "{schema}://{host}{port}/{workflow}/{signature}/?{params}".format(
                schema=self.secure_protocol[0],
                host=self.host,
                port="" if not self.port else f":{self.port}",
                workflow=self.pprint(self.workflow, privacy, safe=""),
                signature=self.pprint(self.signature, privacy, safe=""),
                params=NotifyWorkflows.urlencode(params),
            )
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # store values if provided
        entries = NotifyWorkflows.split_path(results["fullpath"])

        # Display image?
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyWorkflows.template_args["image"]["default"]
            )
        )

        # Support Power Automate URL
        results["power_automate"] = parse_bool(
            results["qsd"].get(
                "powerautomate",
                results["qsd"].get(
                    "pa",
                    NotifyWorkflows.template_args["pa"]["default"]
                )
            )
        )

        # Wrap Text?
        results["wrap"] = parse_bool(
            results["qsd"].get(
                "wrap", NotifyWorkflows.template_args["wrap"]["default"]
            )
        )

        # Template Handling
        if "template" in results["qsd"] and results["qsd"]["template"]:
            results["template"] = NotifyWorkflows.unquote(
                results["qsd"]["template"]
            )

        if "workflow" in results["qsd"] and results["qsd"]["workflow"]:
            results["workflow"] = NotifyWorkflows.unquote(
                results["qsd"]["workflow"]
            )

        elif "id" in results["qsd"] and results["qsd"]["id"]:
            results["workflow"] = NotifyWorkflows.unquote(results["qsd"]["id"])

        else:
            results["workflow"] = (
                None
                if not entries
                else NotifyWorkflows.unquote(entries.pop(0))
            )

        # Signature
        if "signature" in results["qsd"] and results["qsd"]["signature"]:
            results["signature"] = NotifyWorkflows.unquote(
                results["qsd"]["signature"]
            )

        elif "sig" in results["qsd"] and results["qsd"]["sig"]:
            results["signature"] = NotifyWorkflows.unquote(
                results["qsd"]["sig"]
            )

        else:
            # Read information from path
            results["signature"] = (
                None
                if not entries
                else NotifyWorkflows.unquote(entries.pop(0))
            )

        # Version
        if "api-version" in results["qsd"] and results["qsd"]["api-version"]:
            results["version"] = NotifyWorkflows.unquote(
                results["qsd"]["api-version"]
            )

        elif "ver" in results["qsd"] and results["qsd"]["ver"]:
            results["version"] = NotifyWorkflows.unquote(results["qsd"]["ver"])

        # Store our tokens
        results["tokens"] = results["qsd:"]

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support parsing the webhook straight out of workflows
            https://HOST:443/workflows/WORKFLOWID/triggers/manual/paths/invoke
            or
            https://HOST:443/powerautomate/automations/direct/workflows
            /WORKFLOWID/triggers/manual/paths/invoke
        """

        # Match our workflows webhook URL and re-assemble
        result = re.match(
            r"^https?://(?P<host>[A-Z0-9_.-]+)"
            r"(?P<port>:[1-9][0-9]{0,5})?"
            # The new URL structure includes /powerautomate/automations/direct
            r"(?P<power_automate>/powerautomate/automations/direct)?"
            r"/workflows/"
            r"(?P<workflow>[A-Z0-9_-]+)"
            r"/triggers/manual/paths/invoke/?"
            r"(?P<params>\?.+)$",
            url,
            re.I,
        )

        if result:
            # Determine if we're using power automate or not
            power_automate = (
                "&pa=yes"
                if result.group("power_automate")
                else ""
            )

            # Construct our URL
            return NotifyWorkflows.parse_url(
                "{schema}://{host}{port}/{workflow}/{params}{pa}"
                .format(
                    schema=NotifyWorkflows.secure_protocol[0],
                    host=result.group("host"),
                    port=(
                        ""
                        if not result.group("port")
                        else result.group("port")
                    ),
                    workflow=result.group("workflow"),
                    params=result.group("params"),
                    pa=power_automate,
                )
            )
        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

import requests

from .. import exception
from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import NotifyBase


class XMLPayloadField:
    """Identifies the fields available in the JSON Payload."""

    VERSION = "Version"
    TITLE = "Subject"
    MESSAGE = "Message"
    MESSAGETYPE = "MessageType"


# Defines the method to send the notification
METHODS = ("POST", "GET", "DELETE", "PUT", "HEAD", "PATCH")


class NotifyXML(NotifyBase):
    """A wrapper for XML Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "XML"

    # The default protocol
    protocol = "xml"

    # The default secure protocol
    secure_protocol = "xmls"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_Custom_XML"

    # Support attachments
    attachment_support = True

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_128

    # Disable throttle rate for JSON requests since they are normally
    # local anyway
    request_rate_per_sec = 0

    # XSD Information
    xsd_ver = "1.1"
    xsd_default_url = (
        "https://raw.githubusercontent.com/caronc/apprise/master"
        "/apprise/assets/NotifyXML-{version}.xsd"
    )

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{user}@{host}",
        "{schema}://{user}@{host}:{port}",
        "{schema}://{user}:{password}@{host}",
        "{schema}://{user}:{password}@{host}:{port}",
    )

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "method": {
                "name": _("Fetch Method"),
                "type": "choice:string",
                "values": METHODS,
                "default": METHODS[0],
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
        "payload": {
            "name": _("Payload Extras"),
            "prefix": ":",
        },
        "params": {
            "name": _("GET Params"),
            "prefix": "-",
        },
    }

    def __init__(
        self, headers=None, method=None, payload=None, params=None, **kwargs
    ):
        """Initialize XML Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.payload = """<?xml version='1.0' encoding='utf-8'?>
<soapenv:Envelope
    xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <soapenv:Body>
        <Notification{{XSD_URL}}>
            {{CORE}}
            {{ATTACHMENTS}}
       </Notification>
    </soapenv:Body>
</soapenv:Envelope>"""

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = ""

        self.method = (
            self.template_args["method"]["default"]
            if not isinstance(method, str)
            else method.upper()
        )

        if self.method not in METHODS:
            msg = f"The method specified ({method}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # A payload map allows users to over-ride the default mapping if
        # they're detected with the :overide=value.  Normally this would
        # create a new key and assign it the value specified.  However
        # if the key you specify is actually an internally mapped one,
        # then a re-mapping takes place using the value
        self.payload_map = {
            XMLPayloadField.VERSION: XMLPayloadField.VERSION,
            XMLPayloadField.TITLE: XMLPayloadField.TITLE,
            XMLPayloadField.MESSAGE: XMLPayloadField.MESSAGE,
            XMLPayloadField.MESSAGETYPE: XMLPayloadField.MESSAGETYPE,
        }

        self.params = {}
        if params:
            # Store our extra headers
            self.params.update(params)

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        self.payload_overrides = {}
        self.payload_extras = {}
        if payload:
            # Store our extra payload entries (but tidy them up since they will
            # become XML Keys (they can't contain certain characters
            for k, v in payload.items():
                key = re.sub(r"[^A-Za-z0-9_-]*", "", k)
                if not key:
                    self.logger.warning(
                        f"Ignoring invalid XML Stanza element name({k})"
                    )
                    continue

                # Any values set in the payload to alter a system related one
                # alters the system key.  Hence :message=msg maps the 'message'
                # variable that otherwise already contains the payload to be
                # 'msg' instead (containing the payload)
                if key in self.payload_map:
                    self.payload_map[key] = v
                    self.payload_overrides[key] = v

                else:
                    self.payload_extras[key] = v

        # Set our xsd url
        self.xsd_url = (
            None
            if self.payload_overrides or self.payload_extras
            else self.xsd_default_url.format(version=self.xsd_ver)
        )

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform XML Notification."""

        # Prepare HTTP Headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/xml",
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        # Our XML Attachmement subsitution
        xml_attachments = ""

        payload_base = {}

        for key, value in (
            (XMLPayloadField.VERSION, self.xsd_ver),
            (
                XMLPayloadField.TITLE,
                NotifyXML.escape_html(title, whitespace=False),
            ),
            (
                XMLPayloadField.MESSAGE,
                NotifyXML.escape_html(body, whitespace=False),
            ),
            (
                XMLPayloadField.MESSAGETYPE,
                NotifyXML.escape_html(notify_type, whitespace=False),
            ),
        ):

            if not self.payload_map[key]:
                # Do not store element in payload response
                continue
            payload_base[self.payload_map[key]] = value

        # Apply our payload extras
        payload_base.update({
            k: NotifyXML.escape_html(v, whitespace=False)
            for k, v in self.payload_extras.items()
        })

        # Base Entres
        xml_base = "".join(
            [f"<{k}>{v}</{k}>" for k, v in payload_base.items()]
        )

        attachments = []
        if attach and self.attachment_support:
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Custom XML attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    # Prepare our Attachment in Base64
                    entry = '<Attachment filename="{}" mimetype="{}">'.format(
                        NotifyXML.escape_html(
                            (
                                attachment.name
                                if attachment.name
                                else f"file{no:03}.dat"
                            ),
                            whitespace=False,
                        ),
                        NotifyXML.escape_html(
                            attachment.mimetype, whitespace=False
                        ),
                    )
                    entry += attachment.base64()
                    entry += "</Attachment>"
                    attachments.append(entry)

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Custom XML attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Custom XML attachment"
                    f" {attachment.url(privacy=True)}"
                )

            # Update our xml_attachments record:
            xml_attachments = (
                '<Attachments format="base64">'
                + "".join(attachments)
                + "</Attachments>"
            )

        re_map = {
            "{{XSD_URL}}": (
                f' xmlns:xsi="{self.xsd_url}"' if self.xsd_url else ""
            ),
            "{{ATTACHMENTS}}": xml_attachments,
            "{{CORE}}": xml_base,
        }

        # Iterate over above list and store content accordingly
        re_table = re.compile(
            r"(" + "|".join(re_map.keys()) + r")",
            re.IGNORECASE,
        )

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += self.fullpath
        payload = re_table.sub(lambda x: re_map[x.group()], self.payload)

        self.logger.debug(
            f"XML POST URL: {url} (cert_verify={self.verify_certificate!r})"
        )

        self.logger.debug(f"XML Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        if self.method == "GET":
            method = requests.get

        elif self.method == "PUT":
            method = requests.put

        elif self.method == "PATCH":
            method = requests.patch

        elif self.method == "DELETE":
            method = requests.delete

        elif self.method == "HEAD":
            method = requests.head

        else:  # POST
            method = requests.post

        try:
            r = method(
                url,
                data=payload,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code < 200 or r.status_code >= 300:
                # We had a problem
                status_str = NotifyXML.http_response_code_lookup(r.status_code)

                self.logger.warning(
                    "Failed to send JSON %s notification: %s%serror=%s.",
                    self.method,
                    status_str,
                    ", " if status_str else "",
                    str(r.status_code),
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent XML %s notification.", self.method)

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending XML "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
            self.fullpath.rstrip("/"),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "method": self.method,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Append our GET params into our parameters
        params.update({f"-{k}": v for k, v in self.params.items()})

        # Append our payload extra's into our parameters
        params.update({f":{k}": v for k, v in self.payload_extras.items()})
        params.update({f":{k}": v for k, v in self.payload_overrides.items()})

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyXML.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyXML.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}{fullpath}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=(
                NotifyXML.quote(self.fullpath, safe="/")
                if self.fullpath
                else "/"
            ),
            params=NotifyXML.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # store any additional payload extra's defined
        results["payload"] = {
            NotifyXML.unquote(x): NotifyXML.unquote(y)
            for x, y in results["qsd:"].items()
        }

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyXML.unquote(x): NotifyXML.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Add our GET paramters in the event the user wants to pass these along
        results["params"] = {
            NotifyXML.unquote(x): NotifyXML.unquote(y)
            for x, y in results["qsd-"].items()
        }

        # Set method if not otherwise set
        if "method" in results["qsd"] and len(results["qsd"]["method"]):
            results["method"] = NotifyXML.unquote(results["qsd"]["method"])

        return results
